@@ -6,9 +6,7 @@ mod pcap_input;
 
 pub type Time = Duration;
 
-pub use self::config::{
-    EvalConfig, EvaluatorChoice, ExecutionMode, Statistics, TimeFormat, TimeRepresentation, Verbosity,
-};
+pub use self::config::{EvalConfig, ExecutionMode, Statistics, TimeFormat, TimeRepresentation, Verbosity};
 pub use self::io_handler::OutputChannel;
 pub(crate) use self::io_handler::{create_event_source, EventSource, EventSourceConfig, OutputHandler};
 
